try:
    import numpy as np
except ImportError:
    import sys
    import subprocess as sp
    sp.check_call([sys.executable, '-m', 'pip', 'install', 'numpy'])
    import numpy as np
from datetime import datetime


DEBUG = True


class Armijo:
    def __init__(self, loss, alpha=0.5, beta=0.5, dim_x=3, terminate=1e-5, init_x=None, patience_threshold=10):
        assert alpha > 0. and alpha < 1.
        assert beta > 0. and beta < 1.
        self.loss = loss
        self.alpha = alpha
        self.beta = beta
        self.dim_x = dim_x
        self.terminate = terminate
        self.patience_threshold = patience_threshold

        self.step_n = 0
        if init_x is None:
            print(f'[{datetime.now()}] x is randomly initialized.(Dimension: {self.dim_x})')
            self.x = self.init()
        else:
            print(f'[{datetime.now()}] init_x is not None({init_x}). dim_x will be redefined by referring to it.')
            self.dim_x = len(init_x)
            self.x = np.array(init_x)

    def step(self, n_times=10):
        patience = 0
        for _ in range(n_times):
            self.step_n += 1
            if DEBUG: print(f'[step {self.step_n}] x = {self.x}, f(x) = {self.loss(self.x)}')
            step_size = self.calc_step_size(self.x)             # Calculate Step Size
            grad = self.loss.gradient(self.x)                   # Calculate Gradient
            self.x = self.x - step_size * grad                  # Update X
            if np.inner(grad, grad) < self.terminate:
                patience += 1
                if patience == self.patience_threshold:
                    print(f'[{datetime.now()}] Process is terminated.', end='')
                    print('(Gradient is too small to continue. ', end='')
                    print(f'The squre of the norm of gradient was < {self.terminate} for {self.patience_threshold} steps)')
                    break
            else:
                patience = 0
        print(f'[{datetime.now()}] Final Step {self.step_n}')
        print(f'[{datetime.now()}] x = {self.x}, f(x) = {self.loss(self.x)}')

    def calc_step_size(self, x):
        beta_pow = 1.
        grad = self.loss.gradient(x)
        while True:
            if self.loss(self.x - beta_pow * grad) > (self.loss(self.x) - beta_pow * self.alpha * np.inner(grad, grad)):
                if DEBUG: print(f'  [searching step size] step size = {beta_pow}')
                beta_pow *= self.beta
            else:
                if DEBUG: print(f'  [searching step size] step size = {beta_pow}')
                return beta_pow

    def init(self):
        return np.random.rand(self.dim_x)

